/****************************************************************************
** Unit allowing the use of extended memory in Turbo C/C++ programs.       **
**  by Steven H Don                                                        **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************/
#include <dos.h>

int XMSInit ();
void XMSGetMem (unsigned int *Handle, unsigned int Size);
void XMSFreeMem (unsigned int Handle);
void XMSCopy (unsigned int SHandle, long SOfs, unsigned int THandle, long TOfs, long Amount);
void XMSLock (unsigned int Handle, long *LinearAddress);
void XMSUnLock (unsigned int Handle);

//Structure used for copying data
struct EMMS {
  long          Amount;
  unsigned int SHandle;
  long            SOfs;
  unsigned int THandle;
  long            TOfs;
};

//Structure used in calling the XMS manager
struct XMSRegs {
  unsigned int AX, BX, DX, SI, DS;
};

//Pointer to the XMS manager
void far *XMSPtr;

//Registers
unsigned int xAX, xBX, xDX, xSI, xDS;

/****************************************************************************
** Calls the extended memory manager.                                      **
****************************************************************************/
void XMSCall (unsigned int FunctionNumber, struct XMSRegs *XRegs)
{
  unsigned int S, O;
  FunctionNumber <<= 8;
  //Done to prevent "XMSRegs is never used"
  S = FP_SEG (XRegs);
  O = FP_OFF (XRegs);
  asm {
    les di, XMSPtr
    mov ax, FunctionNumber //Retrieve function number

    mov cx, ds             //Copy DS into CX
    push ds                //And store it

    mov di, O              //Retrieve pointer to register values
    mov ds, S
    mov bx, [di+2]         //Retrieve BX
    mov dx, [di+4]         //Retrieve DX
    mov si, [di+6]
    mov ds, [di+8]
    mov es, cx             //The pointer XMSPtr is stored in the original DS, which was copied into CX
    call es:XMSPtr         //The actual call is made here
    mov cx, ds             //The new value of DS must be stored
    mov di, O              //Get the pointer to the registers again
    mov ds, S
    mov [di], ax           //Return registers
    mov [di+2], bx
    mov [di+4], dx
    mov [di+6], si
    mov [di+8], cx
    pop ds                 //Restore DS
  }
}

/****************************************************************************
** Checks to see if an extended memory manager has been installed.         **
** returns true if XMS manager found, false if not.                        **
****************************************************************************/
int XMSInit ()
{
  struct REGPACK Regs;

  //Multiplexer code 4300h - HIMEM.SYS installation status
  Regs.r_ax = 0x4300;
  intr (0x2F, &Regs);
  //Returns AL = 80h if installed
  if ((Regs.r_ax & 0xFF) == 0x80) {
    //Multiplexer code 4310h - Store address for calling XMS functions
    Regs.r_ax = 0x4310;
    intr (0x2F, &Regs);
    XMSPtr = MK_FP (Regs.r_es, Regs.r_bx);
    printf ("%p\n", XMSPtr);
    return 1;
  } else
   return 0;
}

/****************************************************************************
** Allocates an extended memory block of Size K-bytes and returns a handle **
****************************************************************************/
void XMSGetMem (unsigned int *Handle, unsigned int Size)
{
  struct XMSRegs Xr;

  //Set up registers to call XMS manager
  Xr.AX = 0;
  Xr.BX = 0;
  Xr.DX = Size;
  Xr.SI = 0;
  Xr.DS = 0;
  //Call XMS function 9 - allocate extended memory block
  XMSCall (9, &Xr);

  //Error if AX is 0 and BX > 80h
  if (Xr.AX == 0 && Xr.BX >= 128)
    //If so, return 0
    *Handle = 0;
  else
    //Otherwise, return handle
    *Handle = Xr.DX;
}

/****************************************************************************
** Frees an extended memory block allocated by XMSGetMem                   **
****************************************************************************/
void XMSFreeMem (unsigned int Handle)
{
  struct XMSRegs Xr;

  //Set up registers to call XMS manager
  Xr.AX = 0;
  Xr.BX = 0;
  Xr.DX = Handle;
  Xr.SI = 0;
  Xr.DS = 0;
  //Call XMS function 10 - free extended memory block
  XMSCall (10, &Xr);
}

/****************************************************************************
** Copies "amount" bytes of data from a memory block SHandle starting at   **
** offset SOfs within that block to another memory block THandle at offset **
** TOfs within that block. To copy to/from conventional memory, specify    **
** handle number 0 and use a SEGM:OFFS pointer for offset values.          **
*****************************************************************************
** NOTE: This function can only copy EVEN numbers of bytes. A single byte  **
** copy is therefore not possible.                                         **
****************************************************************************/
void XMSCopy (unsigned int SHandle, long SOfs, unsigned int THandle, long TOfs, long Amount)
{
  struct EMMS    Mi;
  struct XMSRegs Xr;

  //Set up structure to pass to the XMS manager
  Mi.Amount = Amount;
  Mi.SHandle = SHandle;
  Mi.SOfs = SOfs;
  Mi.THandle = THandle;
  Mi.TOfs = TOfs;
  Xr.SI = FP_OFF (&Mi);
  Xr.DS = FP_SEG (&Mi);
  //Call XMS function 11 - copy data
  XMSCall (11, &Xr);
}

/****************************************************************************
** This locks the position of the EMB in XMS memory. The XMS manager often **
** moves EMBs around to allow more efficient allocation.                   **
****************************************************************************/
void XMSLock (unsigned int Handle, long *LinearAddress)
{
  struct XMSRegs Xr;

  //Set up registers to call XMS manager
  Xr.AX = 0;
  Xr.BX = 0;
  Xr.DX = Handle;
  Xr.SI = 0;
  Xr.DS = 0;
  //Call XMS function 12 - lock extended memory block
  XMSCall (12, &Xr);
  //Return linear address of EMB
  *LinearAddress = (Xr.DX << 16) | Xr.BX;
}

/****************************************************************************
** This unlocks the position of the EMB in XMS memory.                     **
****************************************************************************/
void XMSUnLock (unsigned int Handle)
{
  struct XMSRegs Xr;

  //Set up registers to call XMS manager
  Xr.AX = 0;
  Xr.BX = 0;
  Xr.DX = Handle;
  Xr.SI = 0;
  Xr.DS = 0;
  //Call XMS function 13 - unlock extended memory block
  XMSCall (13, &Xr);
}